import glob
import numpy as np
from keras.utils import to_categorical


class DataGenerator:
    def __init__(self, data_directory, samples):
        self.data_directory = data_directory
        self.samples = samples
        self.files = set(file_name for file_name in samples)  # <1>
        self.num_samples = None

    def get_num_samples(self, batch_size=128, num_classes=64 * 64):  # <2>
        if self.num_samples is not None:
            return self.num_samples
        else:
            self.num_samples = 0
            for X, y in self._generate(batch_size=batch_size, num_classes=num_classes):
                self.num_samples += X.shape[0]
            return self.num_samples
# <1> Our generator has access to a set of files that we sampled earlier.
# <2> Depending on the application, we may need to know how many examples we have.
# end::data_generator[]

# tag::private_generate[]
    def _generate(self, batch_size, num_classes):
        for zip_file_name in self.files:
            file_name = zip_file_name.replace('.pgn', '') + 'train'
            base = self.data_directory + '/' + file_name + '_features_*.npy'
            for feature_file in glob.glob(base):
                label_file = feature_file.replace('features', 'labels')
                x = np.load(feature_file)
                y = np.load(label_file)
                x = x.astype('float32')
                y = to_categorical(y.astype(int), num_classes)
                while x.shape[0] >= batch_size:
                    x_batch, x = x[:batch_size], x[batch_size:]
                    y_batch, y = y[:batch_size], y[batch_size:]
                    yield x_batch, y_batch  # <1>

# <1> We return or "yield" batches of data as we go.
# end::private_generate[]

# tag::generate[]
    def generate(self, batch_size=40, num_classes=19 * 19):
        while True:
            for item in self._generate(batch_size, num_classes):
                yield item


from __future__ import print_function
from __future__ import absolute_import
import os
import random
import chess 
import chess.pgn
import requests


class Sampler:
    """Sample training and test data from zipped sgf files such that test data is kept stable."""
    def __init__(self, data_dir, num_test_games=100, seed=1337, gamers=['Hikaru','AlexandraBotez','GothamChess','FabianoCaruana', 'DanielNaroditsky', 'Jospem', 'exoticprincess'], month=['01','02''03','04','05','06','07','08','09','10','11','12'], year='2020'):
        self.data_dir = data_dir
        self.num_test_games = num_test_games
        self.test_games = []
        self.train_games = []
        self.test_folder = '/Users/rogermcgrath/Desktop/dlchess/data/test'
        self.gamers = gamers
        self.month = month
        self.year = year
        random.seed(seed)
        self.compute_test_samples()

    def draw_data(self, data_type, num_samples):
        if data_type == 'test':
            return self.test_games
        elif data_type == 'train' and num_samples is not None:
            return self.draw_training_samples(num_samples)
        elif data_type == 'train' and num_samples is None:
            return self.draw_all_training()
        else:
            raise ValueError(data_type + " is not a valid data type, choose from 'train' or 'test'")

    def draw_samples(self, num_sample_games):
        """Draw num_sample_games many training games from index."""
        # self.get_games(self.data_dir,self.gamers, self.month, self.year)
        available_games = []
        for subdir, dirs, files in os.walk(self.data_dir):
            for file in files:
                available_games.append(file)
        print('>>> Total number of games used: ' + str(len(available_games)))
        
        sample_set = set()
        while len(sample_set) < num_sample_games:
            sample = random.choice(available_games)
            if sample not in sample_set:
                sample_set.add(sample)
        print('Drawn ' + str(num_sample_games) + ' samples:')
        return list(sample_set)

    def draw_training_games(self):
        """Get list of all non-test games, that are no later than dec 2014
        Ignore games after cap_year to keep training data stable
        """
        available_games = []
        self.get_games(self.data_dir,self.gamers, self.month, self.year)
        print(self.data_dir)
        for subdir, dirs, files in os.walk(self.data_dir):
            for file in files:
                available_games.append(file)
        
        sample_set = set()
        while len(sample_set) < len(available_games):
            sample = random.choice(available_games)
            if sample not in sample_set:
                self.train_games.append(sample)
        print('Drawn ' + str(len(self.train_games)) + ' samples:')

    def compute_test_samples(self):
        """If not already existing, create local file to store fixed set of test samples"""
        for subdir, dirs, files in os.walk(self.test_folder):
            for file in files:
                self.test_games.append(file)

    def draw_training_samples(self, num_sample_games):
        """Draw training games, not overlapping with any of the test games."""
        available_games = []
        self.get_games(self.data_dir,self.gamers, self.month, self.year)
        for subdir, dirs, files in os.walk(self.data_dir):
            for file in files:
                available_games.append(file)
        
        sample_set = set()
        for i in range(num_sample_games):
            sample = random.choice(available_games)
            if sample not in sample_set:
                self.train_games.append(sample)
        print('Drawn ' + str(len(self.train_games)) + ' samples:')
        return(self.train_games)

    def draw_all_training(self):
        """Draw all available training games."""
        available_games = []
        self.get_games(self.data_dir,self.gamers, self.month, self.year)
        for subdir, dirs, files in os.walk(self.data_dir):
            for file in files:
                available_games.append(file)
        
        sample_set = set()
        for samples in available_games:
            sample = random.choice(available_games)
            if sample not in sample_set:
                self.train_games.append(sample)
        print('Drawn ' + str(len(self.train_games)) + ' samples:')

    def get_games(self, data_directory, gamers, months, year):
        if data_directory == '/Users/rogermcgrath/Desktop/dlchess/data/training':
            for gamer in gamers:
                for month in months:
                    link = 'https://api.chess.com/pub/player/' + gamer + '/' + 'games' + '/' + year + '/' + month + '/' + 'pgn'
                    pgns = requests.get(link)
                    print(data_directory + '/' + gamer + '_' + month + '.pgn')
                    write = open(data_directory + '/' + gamer + '_' + month + '.pgn', 'w')
                    write.write(pgns.text)
                    pgn = open(data_directory + '/' + gamer + '_' + month + '.pgn')
                    coutner = 0
                    while True:
                        headers = chess.pgn.read_headers(pgn)
                        if headers is None:
                            break
                        # elif gamer == 'Hikaru' and coutner == 122 or coutner == 41 or coutner == 34 or coutner == 33 or coutner == 32:
                        #     coutner += 1
                        #     break
                        file = open(data_directory + '/' + gamer + '_' + month + str(coutner) + '.pgn', 'w')
                        file.write(str(chess.pgn.read_game(pgn)))
                        coutner += 1
                    os.remove(pgn)
                gamer = ['Carlsen', 'Morphy']
                for game in gamer:
                    pgn = open(data_directory + '/' + game + '.pgn')
                    print(str(pgn))
                    coutner = 0 
                    while True:
                        headers = chess.pgn.read_headers(pgn)
                        if headers is None:
                            break
                        file = open(data_directory + '/' + game + '_' + str(coutner) + '.pgn', 'w')
                        file.write(str(chess.pgn.read_game(pgn)))
                        coutner += 1
                    os.remove(pgn)
        else:
            gamer = ['Caruana','Grischuk','Huebner','Kasparov','Nakamura']
            for game in gamer:
                pgn = open(data_directory + '/' + game + '.pgn')
                print(str(pgn))
                coutner = 0 
                while True:
                    headers = chess.pgn.read_headers(pgn)
                    if headers is None:
                        break
                    file = open(data_directory + '/' + game + '_' + str(coutner) + '.pgn', 'w')
                    file.write(str(chess.pgn.read_game(pgn)))
                    coutner += 1
                os.remove(pgn)




from __future__ import absolute_import

# tag::base_imports[]
import os.path
import tarfile
import gzip
import glob
import shutil
import sys
import chess
import chess.pgn

import numpy as np
from keras.utils import to_categorical
# end::base_imports[]
from sample import Sampler
from generator import DataGenerator
sys.path.append('/Users/rogermcgrath/Desktop/dlchess/encoders')
from oneplane import OnePlaneEncoder

class ChessDataProcessor:
    def __init__(self, encoder='oneplane', data_directory='data'):
        self.encoder = OnePlaneEncoder((8,8))
        self.data_dir = data_directory
# end::processor_init[]

# tag::load_go_data[]
    def load_chess_data(self, data_type='train', num_samples=1000,
                     use_generator=False):

        sampler = Sampler(data_dir=self.data_dir)
        data = sampler.draw_data(data_type, num_samples)

        zip_names = set()
        for filename in data:
            if not filename == '.DS_Store':
                zip_names.add(filename)  # <5>
        for zip_name in zip_names:
            base_name = zip_name.replace('.pgn', '')
            data_file_name = base_name + data_type
            if not os.path.isfile(self.data_dir + '/' + data_file_name):
                self.process_zip(zip_name, data_file_name)
        if use_generator:
            generator = DataGenerator(self.data_dir, data)
            return generator  # <2>
        else:
            features_and_labels = self.consolidate_games(data_type, data)
            return features_and_labels  # <3>

# <1> As `data_type` you can choose either 'train' or 'test'.
# <2> `num_samples` refers to the number of games to load data from.
# <3> We download all games from KGS to our local data directory. If data is available, it won't be downloaded again.
# <4> The `Sampler` instance selects the specified number of games for a data type.
# <5> We collect all zip file names contained in the data in a list.
# <6> Then we group all SGF file indices by zip file name.
# <7> The zip files are then processed individually.
# <8> Features and labels from each zip are then aggregated and returned.
# end::load_go_data[]

# tag::unzip_data[]

# <1> Unpack the `gz` file into a `tar` file.
# <2> Remove ".gz" at the end to get the name of the tar file.
# <3> Copy the contents of the unpacked file into the `tar` file.
# end::unzip_data[]

# tag::read_sgf_files[]
    def process_zip(self, zip_file_name, data_file_name):
        total_examples = self.num_total_examples(zip_file_name)  # <1>
        shape = self.encoder.shape()  # <2>
        feature_shape = np.insert(shape, 0, np.asarray([total_examples]))
        features = np.zeros(feature_shape)
        labels = np.zeros((total_examples,))

        counter = 0
        pgn = open(self.data_dir + '/' + zip_file_name)
        game = chess.pgn.read_game(pgn)
        board = chess.Board()
        for move in game.mainline_moves():
            if counter == total_examples:
                break
            board.push(move)
            features[counter] = self.encoder.encode(board)
            labels[counter] = self.encoder.encode_point(move)
            counter += 1

        feature_file_base = self.data_dir + '/' + data_file_name + '_features_%d'
        label_file_base = self.data_dir + '/' + data_file_name + '_labels_%d'

        chunk = 0  # Due to files with large content, split up after chunksize
        chunksize = 40
        while features.shape[0] >= chunksize:  # <1>
            feature_file = feature_file_base % chunk
            label_file = label_file_base % chunk
            chunk += 1
            current_features, features = features[:chunksize], features[chunksize:]
            current_labels, labels = labels[:chunksize], labels[chunksize:]  # <2>
            np.save(feature_file, current_features)
            np.save(label_file, current_labels)  # <3>
# <1> We process features and labels in chunks of size 1024.
# <2> The current chunk is cut off from features and labels...
# <3> ...  and then stored in a separate file.
# end::store_features_and_labels[]

# tag::consolidate_games[]
    def consolidate_games(self, data_type, samples):
        files_needed = set(file_name for file_name in samples)
        file_names = []
        for zip_file_name in files_needed:
            file_name = zip_file_name.replace('.pgn', '') + data_type
            file_names.append(file_name)

        feature_list = []
        label_list = []
        for file_name in file_names:
            file_prefix = file_name.replace('.pgn', '')
            base = self.data_dir + '/' + file_prefix + '_features_*.npy'
            for feature_file in glob.glob(base):
                label_file = feature_file.replace('features', 'labels')
                x = np.load(feature_file)
                y = np.load(label_file)
                x = x.astype('float32')
                y = to_categorical(y.astype(int), 64 * 64)
                feature_list.append(x)
                label_list.append(y)
        features = np.concatenate(feature_list, axis=0)
        labels = np.concatenate(label_list, axis=0)
        np.save('{}/features_{}.npy'.format(self.data_dir, data_type), features)
        np.save('{}/labels_{}.npy'.format(self.data_dir, data_type), labels)

        return features, labels

    def num_total_examples(self, zip_file):
        total_examples = 0
        pgn = open(self.data_dir + '/' + zip_file)
        game = chess.pgn.read_game(pgn)
        board = chess.Board()
        num_moves = 0
        for move in game.mainline_moves():
            if move in board.pseudo_legal_moves:
                board.push(move)
                num_moves += 1
            else:
                break
        total_examples = total_examples + num_moves
        return num_moves


import importlib
# end::importlib[]

__all__ = [
    'Encoder',
    'get_encoder_by_name',
]


# tag::base_encoder[]
class Encoder:
    def name(self):  # <1>
        raise NotImplementedError()

    def encode(self, board):  # <2>
        raise NotImplementedError()

    def encode_point(self, point):  # <3>
        raise NotImplementedError()

    def decode_point_index(self, index):  # <4>
        raise NotImplementedError()

    def num_points(self):  # <5>
        raise NotImplementedError()

    def shape(self):  # <6>
        raise NotImplementedError()

# <1> Lets us support logging or saving the name of the encoder our model is using.
# <2> Turn a Go board into a numeric data.
# <3> Turn a Go board point into an integer index.
# <4> Turn an integer index back into a Go board point.
# <5> Number of points on the board, i.e. board width times board height.
# <6> Shape of the encoded board structure.
# end::base_encoder[]


# tag::encoder_by_name[]
def get_encoder_by_name(name, board_size):  # <1>
    if isinstance(board_size, int):
        board_size = (board_size, board_size)  # <2>
    module = importlib.import_module('dlchess.encoders.' + name)
    constructor = getattr(module, 'create')  # <3>
    return constructor(board_size)

# <1> We can create encoder instances by referencing their name.
# <2> If board_size is one integer, we create a square board from it.
# <3> Each encoder implementation will have to provide a "create" function that provides an instance.
# end::encoder_by_name[]



from setuptools import setup, find_packages

classifiers = [
    'License :: OSI Approved :: MIT License'
]

setup(
    name='dlchessrog',
    version='0.0.2',
    description = 'My chess bot modules',
    # py_modules=["processor","generator","sample","base","oneplane","layers"],
    # package_dir={'':'src'},
    packages = find_packages(),
    license='MIT'
)


import numpy as np 
import chess 
import chess.svg
from base import Encoder


class OnePlaneEncoder(Encoder):
    def __init__(self, board_size):
        self.board_width, self.board_height = board_size
        self.num_planes = 1

    def name(self):  # <1>
        return 'oneplane'

    def encode(self, board):  # <2>
        board_matrix = np.zeros(self.shape())
        counter = 0
        for r in range(self.board_height):
            for c in range(self.board_width):
                if counter < 64:
                    p = board.piece_at(counter)
                    p = str(p)
                    color = board.color_at(counter)
                    if color is True:
                        if p == 'P':
                            board_matrix[0,r,c] = 1
                        elif p == 'B':
                            board_matrix[0,r,c] = 2
                        elif p == 'N':
                            board_matrix[0,r,c] = 3
                        elif p == 'R':
                            board_matrix[0,r,c] = 4
                        elif p == 'Q':
                            board_matrix[0,r,c] = 5
                        elif p == 'K':
                            board_matrix[0,r,c] = 6
                    else:
                        if p == 'p':
                            board_matrix[0,r,c] = -1
                        elif p == 'b':
                            board_matrix[0,r,c] = -2
                        elif p == 'n':
                            board_matrix[0,r,c] = -3
                        elif p == 'r':
                            board_matrix[0,r,c] = -4
                        elif p == 'q':
                            board_matrix[0,r,c] = -5
                        elif p == 'k':
                            board_matrix[0,r,c] = -6
                    counter += 1
        return board_matrix

# <1> We can reference this encoder by the name "oneplane".
# <2> To encode, we fill a matrix with 1 if the point contains one of the current player's stones, -1 if the point contains the opponent's stones and 0 if the point is empty.
# end::oneplane_encoder[]

# tag::oneplane_encoder_2[]
    def encode_point(self, point):  # <1>
        return 64 * (chess.square(chess.square_file(point.from_square), chess.square_rank(point.from_square))) + (chess.square(chess.square_file(point.to_square), chess.square_rank(point.to_square)))

    def decode_point_index(self, index):  # <2>
        froms = index // 64
        to = index % 64
        return (froms, to)

    def num_points(self):
        return self.board_width * self.board_height

    def shape(self):
        return self.num_planes, self.board_height, self.board_width

# <1> Turn a board point into an integer index.
# <2> Turn an integer index into a board point.
# end::oneplane_encoder_2[]


# tag::oneplane_create[]
def create(board_size):
    return OnePlaneEncoder(board_size)
# end::oneplane_create[]